/*--------------------------------------------------------------------
 * $Id: wcloud3d.c 2623 2011-12-23 10:52:38Z robert.buras $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2012 by Arve Kylling, Bernhard Mayer,
 *                            Claudia Emde, Robert Buras
 *
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <math.h>
#include <limits.h>
#include <string.h>
#include <float.h>
#include "ascii.h"
#include "wcloud3d.h"

#if HAVE_LIBNETCDF
#include <netcdf.h>
#endif

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif

/* Maximum number of characters per line */
#define MAX_LENGTH_OF_LINE     65536


/*******************************************************/
/* Read data from a 3D cloud description file.         */
/*******************************************************/
 
int read_3D_caoth ( char      *filename, 
		    int       *Nx,
		    int       *Ny,
		    int       *Nz,
		    double    *delX,
		    double    *delY,
		    float    **z, 
		    float  ****lwc,
		    float  ****reff,
		    float  ****ext,
		    float  ****g1, 
		    float  ****g2, 
		    float  ****ff, 
		    float  ****f, 
		    float  ****ssa, 
		    float  ****dscale, 
		    float     *rmin,
		    float     *rmax,
		    int       *cldproperties,
		    int      **threed,
		    int        quiet )
{
  int i=0, number=0, status=0;
  int ix=0, iy=0, iz=0;

  int rows=0, min_columns=0, max_columns=0, max_length=0;

  char line[MAX_LENGTH_OF_LINE+1]="";
  char *string=NULL;

  FILE *file=NULL;
  char **array=NULL;
  char *dummy=NULL;

  /* check input file */
  status =  ASCII_checkfile (filename, 
			     &rows,
			     &min_columns,
			     &max_columns,
			     &max_length);
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n",
	     status, filename);
    return status;
  }
    
  if (rows<2) {
    fprintf (stderr, "Error: found less than two rows in %s\n", filename);
    return -1;
  }

  if (min_columns<3) {
    fprintf (stderr, "Error: found less than three columns in %s\n", filename);
    return -1;
  }
  
  /* reset string to the beginning of the memory block */
  string = line;
  
  /* open file */
  if ( (file = fopen(filename, "r")) == NULL)  
    return ASCIIFILE_NOT_FOUND;
  

  /* first line */
  number=0;
  while (number==0) {
    fgets (string, MAX_LENGTH_OF_LINE, file);
    status = ASCII_parsestring (string, &array, &number);

    if (status!=0) {
      fprintf (stderr, "Error %d reading 1st line of %s\n", status, filename);
      return status;
    }
  }

  if (number<3) {
    fprintf (stderr, "Error: expected at least 3 values in 1st line, found only %d\n", number);
    return -1;
  }
  
  /* first 3 values are number of cells */
  *Nx = strtol (array[0], &dummy, 0);
  *Ny = strtol (array[1], &dummy, 0);
  *Nz = strtol (array[2], &dummy, 0);

  if (number>3)
    *cldproperties = strtol (array[3], &dummy, 0);
    
  if (number>4) {
    fprintf (stderr, "*** WARNING: Found a 5th number in the first line of %s.\n", filename);
    fprintf (stderr, "*** Possibly you use the old format. Please be aware that wspec\n");
    fprintf (stderr, "*** doesn't exist anymore and the 5th number is ignored!\n");
  }
  
  free (array);

  if (!quiet) {
    fprintf (stderr, " ... reading %d x %d x %d data points from %s\n",
	     *Nx, *Ny, *Nz, filename);
    fprintf (stderr, " ... cldproperties = %d\n", *cldproperties);
  }

  /* 2nd line */
  number=0;
  while (number==0) {
    fgets (string, MAX_LENGTH_OF_LINE, file);
    status = ASCII_parsestring (string, &array, &number);

    if (status!=0) {
      fprintf (stderr, "Error %d reading 1st line of %s\n", status, filename);
      return status;
    }
  }

  if (number!=*Nz+1+2) {
    fprintf (stderr, "Error: found %d z levels, expected %d\n",
	     number-2, *Nz+1);
    return -1;
  }

  /* horizontal grid sizes, in km */
  *delX = strtod (array[0], &dummy);
  *delY = strtod (array[1], &dummy);

  if (!quiet) {
    fprintf (stderr, " ... cloud grid size in x direction: %g km\n", *delX);
    fprintf (stderr, " ... cloud grid size in y direction: %g km\n", *delY);
  }

  /* allocate memory for altitude levels */
  *z = calloc ((size_t) *Nz+1, sizeof(float));
  if (*z==NULL) {
    fprintf (stderr, "Error allocating memory for z\n");
    return -1;
  }
	     
  *threed = calloc ((size_t) *Nz, sizeof(int));
  if (*threed==NULL) {
    fprintf (stderr, "Error allocating memory for threed\n");
    return -1;
  }
	     
  /* set altitude levels and convert from km to m */
  for (i=2; i<=*Nz+2; i++)
    (*z)[i-2] = (float) strtod(array[i], &dummy);

  free(array);

  /* allocate memory for optical depth, asymmetry factor, and single scattering albedo */
  *ext    = calloc((size_t) *Nz, sizeof(float **));
  if (*ext==NULL) {
    fprintf (stderr, "Error allocating memory for ext\n");
    return -1;
  }
  *g1     = calloc((size_t) *Nz, sizeof(float **));
  if (*g1==NULL) {
    fprintf (stderr, "Error allocating memory for g1\n");
    return -1;
  }
  *g2     = calloc((size_t) *Nz, sizeof(float **));
  if (*g2==NULL) {
    fprintf (stderr, "Error allocating memory for g2\n");
    return -1;
  }
  *ff     = calloc((size_t) *Nz, sizeof(float **));
  if (*ff==NULL) {
    fprintf (stderr, "Error allocating memory for ff\n");
    return -1;
  }
  *ssa    = calloc((size_t) *Nz, sizeof(float **));
  if (*ssa==NULL) {
    fprintf (stderr, "Error allocating memory for ssa\n");
    return -1;
  }
  *f      = calloc((size_t) *Nz, sizeof(float **));
  if (*f==NULL) {
    fprintf (stderr, "Error allocating memory for f\n");
    return -1;
  }
  *dscale = calloc((size_t) *Nz, sizeof(float **));
  if (*dscale==NULL) {
    fprintf (stderr, "Error allocating memory for dscale\n");
    return -1;
  }
  *lwc    = calloc((size_t) *Nz, sizeof(float **));
  if (*lwc==NULL) {
    fprintf (stderr, "Error allocating memory for lwc\n");
    return -1;
  }
  *reff   = calloc((size_t) *Nz, sizeof(float **));
  if (*reff==NULL) {
    fprintf (stderr, "Error allocating memory for reff\n");
    return -1;
  }
  
  for (iz=0; iz<*Nz; iz++) {
    (*ext)    [iz] = calloc((size_t) *Nx, sizeof(float *));
    if ((*ext)[iz]==NULL) {
      fprintf (stderr, "Error allocating memory for ext[%d]\n",iz);
      return -1;
    }
    (*g1)     [iz] = calloc((size_t) *Nx, sizeof(float *));
    if ((*g1)[iz]==NULL) {
      fprintf (stderr, "Error allocating memory for g1[%d]\n",iz);
      return -1;
    }
    (*g2)     [iz] = calloc((size_t) *Nx, sizeof(float *));
    if ((*g2)[iz]==NULL) {
      fprintf (stderr, "Error allocating memory for g2[%d]\n",iz);
      return -1;
    }
    (*ff)     [iz] = calloc((size_t) *Nx, sizeof(float *));
    if ((*ff)[iz]==NULL) {
      fprintf (stderr, "Error allocating memory for ff[%d]\n",iz);
      return -1;
    }
    (*ssa)    [iz] = calloc((size_t) *Nx, sizeof(float *));
    if ((*ssa)[iz]==NULL) {
      fprintf (stderr, "Error allocating memory for ssa[%d]\n",iz);
      return -1;
    }
    (*f)      [iz] = calloc((size_t) *Nx, sizeof(float *));
    if ((*f)[iz]==NULL) {
      fprintf (stderr, "Error allocating memory for g[%d]\n",iz);
      return -1;
    }
    (*dscale) [iz] = calloc((size_t) *Nx, sizeof(float *));
    if ((*dscale)[iz]==NULL) {
      fprintf (stderr, "Error allocating memory for dscale[%d]\n",iz);
      return -1;
    }
    (*lwc)    [iz] = calloc((size_t) *Nx, sizeof(float *));
    if ((*lwc)[iz]==NULL) {
      fprintf (stderr, "Error allocating memory for lwc[%d]\n",iz);
      return -1;
    }
    (*reff)   [iz] = calloc((size_t) *Nx, sizeof(float *));
    if ((*reff)[iz]==NULL) {
      fprintf (stderr, "Error allocating memory for reff[%d]\n",iz);
      return -1;
    }

    for (ix=0; ix<*Nx; ix++) {
      (*ext)    [iz][ix] = calloc((size_t) *Ny, sizeof(float)); 
      if ((*ext)[iz][ix]==NULL) {
	fprintf (stderr, "Error allocating memory for ext[%d][%d]\n",iz,ix);
	return -1;
      }
      (*g1)     [iz][ix] = calloc((size_t) *Ny, sizeof(float)); 
      if ((*g1)[iz][ix]==NULL) {
	fprintf (stderr, "Error allocating memory for g1[%d][%d]\n",iz,ix);
	return -1;
      }
      (*g2)     [iz][ix] = calloc((size_t) *Ny, sizeof(float)); 
      if ((*g2)[iz][ix]==NULL) {
	fprintf (stderr, "Error allocating memory for g2[%d][%d]\n",iz,ix);
	return -1;
      }
      (*ff)     [iz][ix] = calloc((size_t) *Ny, sizeof(float)); 
      if ((*ff)[iz][ix]==NULL) {
	fprintf (stderr, "Error allocating memory for ff[%d][%d]\n",iz,ix);
	return -1;
      }
      (*ssa)    [iz][ix] = calloc((size_t) *Ny, sizeof(float)); 
      if ((*ssa)[iz][ix]==NULL) {
	fprintf (stderr, "Error allocating memory for ssa[%d][%d]\n",iz,ix);
	return -1;
      }
      (*f)      [iz][ix] = calloc((size_t) *Ny, sizeof(float)); 
      if ((*f)[iz][ix]==NULL) {
	fprintf (stderr, "Error allocating memory for f[%d][%d]\n",iz,ix);
	return -1;
      }
      (*dscale) [iz][ix] = calloc((size_t) *Ny, sizeof(float)); 
      if ((*dscale)[iz][ix]==NULL) {
	fprintf (stderr, "Error allocating memory for dscale[%d][%d]\n",iz,ix);
	return -1;
      }
      (*lwc)    [iz][ix] = calloc((size_t) *Ny, sizeof(float));  
      if ((*lwc)[iz][ix]==NULL) {
	fprintf (stderr, "Error allocating memory for lwc[%d][%d]\n",iz,ix);
	return -1;
      }
      (*reff)   [iz][ix] = calloc((size_t) *Ny, sizeof(float));  
      if ((*reff)[iz][ix]==NULL) {
	fprintf (stderr, "Error allocating memory for reff[%d][%d]\n",iz,ix);
	return -1;
      }

	/* initialize forward HG fraction with 1 */
	for (iy=0; iy<*Ny; iy++)
	  (*ff)[iz][ix][iy] = 1.0; 
      }
  }


  /* initialize rmin, rmax */
  switch (*cldproperties) {
  case CLD_OPTPROP:
    *rmin=0; *rmax=0;  /* rmin, rmax not used */
    break;

  case CLD_EXTREFF:
  case CLD_LWCREFF:
    *rmin=+FLT_MAX;
    *rmax=-FLT_MAX;
    break;

  default:
    fprintf (stderr, "Error, unknown cloud properties %d in %s\n", *cldproperties, filename);
    return -1;
  }


  for (i=2; i<rows; i++) {
    number=0;
    while (number==0) {
      fgets (string, MAX_LENGTH_OF_LINE, file);
      status = ASCII_parsestring (string, &array, &number);
    
      if (status!=0) {
	fprintf (stderr, "Error %d reading 1st line of %s\n", status, filename);
	return status;
      }
    }

    switch (*cldproperties) {
    case CLD_OPTPROP:
      if (number<6) {
	fprintf (stderr, "Error: found %d columns in row %d, expected 6\n", number, i);
	return -1;
      }
      break;

    case CLD_EXTREFF:
      if (number<5) {
	fprintf (stderr, "Error: found %d columns in row %d, expected 5\n", number, i);
	return -1;
      }
      break;

    case CLD_LWCREFF:
      if (number<5) {
	fprintf (stderr, "Error: found %d columns in row %d, expected 5\n", number, i);
	return -1;
      }
      break;

    default:
      fprintf (stderr, "Error, unknown cloud properties %d in %s\n", *cldproperties, filename);
      return -1;
    }
      
      
    /* cell indices */
    ix = strtol(array[0], &dummy, 0) - 1;
    iy = strtol(array[1], &dummy, 0) - 1;
    iz = strtol(array[2], &dummy, 0) - 1;

    if (ix<0 || ix>=*Nx) {
      fprintf (stderr, "Error, ix = %d out of bounds in line %d\n", ix+1, i+1);
      return -1;
    }

    if (iy<0 || iy>=*Ny) {
      fprintf (stderr, "Error, iy = %d out of bounds in line %d\n", iy+1, i+1);
      return -1;
    }

    if (iz<0 || iz>=*Nz) {
      fprintf (stderr, "Error, iz = %d out of bounds in line %d\n", iz+1, i+1);
      return -1;
    }


    /* copy data to result arrays */
    switch (*cldproperties) {
    case CLD_OPTPROP:
      (*ext)[iz][ix][iy] = strtod (array[3], &dummy)   / 1000.0;  /* convert from km-1 to m-1 */
      (*g1) [iz][ix][iy] = strtod (array[4], &dummy);
      (*g2) [iz][ix][iy] = 0;
      (*ff) [iz][ix][iy] = 1.0;
      (*ssa)[iz][ix][iy] = strtod (array[5], &dummy);
      break;

    case CLD_EXTREFF:
      (*ext) [iz][ix][iy] = strtod (array[3], &dummy) / 1000.0;  /* convert from km-1 to m-1 */
      (*reff)[iz][ix][iy] = strtod (array[4], &dummy);
      
      if ((*reff)[iz][ix][iy] < *rmin)  *rmin = (*reff)[iz][ix][iy];
      if ((*reff)[iz][ix][iy] > *rmax)  *rmax = (*reff)[iz][ix][iy];
  
      break;

    case CLD_LWCREFF:
      (*lwc) [iz][ix][iy] = strtod (array[3], &dummy);
      (*reff)[iz][ix][iy] = strtod (array[4], &dummy);

      if ((*reff)[iz][ix][iy] < *rmin)  *rmin = (*reff)[iz][ix][iy];
      if ((*reff)[iz][ix][iy] > *rmax)  *rmax = (*reff)[iz][ix][iy];
      break;

    default:
      fprintf (stderr, "Error, unknown cloud properties %d in %s\n", *cldproperties, filename);
      return -1;
    }

    /* set flag indicating that this layer is 3D */
    (*threed)[iz] = 1;

    free(array);
  }


  /* close input file */
  fclose (file);

  if (!quiet)
    fprintf (stderr, " ... read %d data points from %s\n", 
	     rows-2, filename);

  return 0;
}





/*******************************************************/
/* Read data from a 2D albedo description file.        */
/*******************************************************/

int read_2D_albedo (char *filename, 
		    int *Nx, int *Ny,
		    double *delX, double *delY,
		    double ***albedo,
		    int quiet)
{
  int i=0, status=0;
  int ix=0, iy=0;

  double **value=NULL;
  int rows=0, min_columns=0, max_columns=0;

#if HAVE_LIBNETCDF
  struct stat buf;
  struct tm *tmstruct=NULL;

  size_t start[1] = {0};
 
  size_t tstart[2] = {0,0};
  size_t tcount[2] = {0,0};
 
  size_t n=0;

  int    ncid   =0, idd_nx =0, idd_ny =0;
  int    id_delX=0, id_delY=0, id_type=0;

  float *temp=NULL;
#endif


  /* try to open as netCDF file */
  
#if HAVE_LIBNETCDF
  status = nc_open (filename, NC_NOWRITE, &ncid);
  if (status==NC_NOERR) {

    if (!quiet)
      fprintf (stderr, " ... reading 2D albedo data from netCDF file %s\n", filename);

    /* determine file date and stop if the file was created before */
    /* September 7, 2007 where the format was changed              */
    stat (filename, &buf);
    
    tmstruct=gmtime(&buf.st_mtime);
    
    if (buf.st_mtime<1189187082) {
      fprintf (stderr, "\n");
      fprintf (stderr, "*** Error %s was last changed %s", filename, ctime(&buf.st_mtime));
      fprintf (stderr, "*** and the MYSTIC albedo2D convention has been changed on\n");
      fprintf (stderr, "*** September 7, 2007. It is likely that you use the \n");
      fprintf (stderr, "*** old convention! Please check the documentation!\n");
      fprintf (stderr, "\n");
      
      return -1;
    }
    
    fprintf (stderr, "\n");
    fprintf (stderr, "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
    fprintf (stderr, "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
    fprintf (stderr, "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
    fprintf (stderr, "!!! Careful! The netcdf file format for 2D albedo was !!!\n");
    fprintf (stderr, "!!! changed on September 7, 2007. In particular,      !!!\n");
    fprintf (stderr, "!!! x and y were interchanged so that an albedo file  !!!\n");
    fprintf (stderr, "!!! viewed with ncview should look like a map now,    !!!\n");
    fprintf (stderr, "!!! with (x,y)=(1,1) in the lower left corner.        !!!\n");
    fprintf (stderr, "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
    fprintf (stderr, "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
    fprintf (stderr, "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
    fprintf (stderr, "\n");
    
    /* get dimension id for "nx" */
    status = nc_inq_dimid (ncid, "nx", &idd_nx);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading Nx from %s\n", status, filename);
      return status;
    }
    
    /* get dimension length for "nx" */
    status = nc_inq_dimlen (ncid, idd_nx, &n);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading Nx from %s\n", status, filename);
      return status;
    }

    *Nx=n;

    /* get dimension id for "ny" */
    status = nc_inq_dimid (ncid, "ny", &idd_ny);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading Ny from %s\n", status, filename);
      return status;
    }
    
    /* get dimension length for "ny" */
    status = nc_inq_dimlen (ncid, idd_ny, &n);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading Ny from %s\n", status, filename);
      return status;
    }

    *Ny = n;

    /* get variable id for "delX" */
    status = nc_inq_varid (ncid, "dx", &id_delX);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading delX from %s\n", 
	       status, filename);
      return status;
    }

    /* read "delX" */
    status = nc_get_var1_double (ncid, id_delX, start, delX);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading delX from %s\n", 
	       status, filename);
      return status;
    }


    /* get variable id for "delY" */
    status = nc_inq_varid (ncid, "dy", &id_delY);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading dy from %s\n", 
	       status, filename);
      return status;
    }

    /* read "delY" */
    status = nc_get_var1_double (ncid, id_delY, start, delY);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading dy from %s\n", 
	       status, filename);
      return status;
    }


    *delX *= 1000.0;
    *delY *= 1000.0;


    /* allocate memory for 2D albedo field */
    *albedo = calloc((size_t) *Nx, sizeof(double *));

    for (ix=0; ix<*Nx; ix++)
      (*albedo)[ix] = calloc((size_t) *Ny, sizeof(double));
    

    /* read data */

    /* get variable id for "albedo" */
    status = nc_inq_varid (ncid, "albedo", &id_type);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading albedo from %s\n", status, filename);
      return status;
    }

    temp = calloc (*Nx, sizeof(float));
    
    /* read type */
    tstart[1] = 0;   /* start with first row          */
    tcount[1] = *Nx; /* read *Nx elements for each iy */

    tcount[0] = 1;  /* read one line at a time */

    for (iy=0; iy<*Ny; iy++) {
      tstart[0] = iy;

      status = nc_get_vara_float (ncid, id_type, tstart, tcount, temp);
      if (status!=NC_NOERR) {
	fprintf (stderr, "Error %d reading albedo from %s\n", status, filename);
	return status;
      }

      for (ix=0; ix<*Nx; ix++)
	(*albedo)[ix][iy] = temp[ix];
    }
    nc_close (ncid);

    free(temp);
    
    if (!quiet)
      fprintf (stderr, " ... read %d data points from %s\n", 
	       (*Nx)*(*Ny), filename);
  }
  else {
    if (!quiet)
      fprintf (stderr, " ... %s not in netCDF format, trying to open as ASCII\n", 
	       filename);
#endif

    /* read file; ASCII_file2double is a waste of memory in this */
    /* case, but it is very convenient because comments and      */
    /* everything are handled correctly.                         */
    
    status = ASCII_file2double (filename,
				&rows, &max_columns, &min_columns, 
				&value);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading file %s\n", status, filename);
      return status;
    }
    
    if (max_columns<4) {
      fprintf (stderr, "Error, found less than four columns in %s\n", 
	       filename);
      return -1;
    }
    
    if (rows<1) {
      if (!quiet)
	fprintf (stderr, "Error, no header in %s\n", filename);
      return -1;
  }
    
    
    /* 1st line, number of cells */
    *Nx = (int) (value[0][0] + 0.5);
    *Ny = (int) (value[0][1] + 0.5);
    
    
    /* 1st line, horizontal distances, convert from km to m */
    *delX = value[0][2] * 1000.0;
    *delY = value[0][3] * 1000.0;
    
    if (!quiet) {
      fprintf (stderr, " ... x-distance: %g\n", *delX);
      fprintf (stderr, " ... y-distance: %g\n", *delY);
    }
    
    /* allocate memory for albedo */
    *albedo = calloc((size_t) *Nx, sizeof(double *));
    
    for (ix=0; ix<*Nx; ix++)
      (*albedo)[ix] = calloc((size_t) *Ny, sizeof(double));
    
    
    for (i=1; i<rows; i++) {
      ix = (int) (value[i][0] + 0.5) - 1;
      iy = (int) (value[i][1] + 0.5) - 1;
      
      if (ix<0 || ix>=*Nx) {
	fprintf (stderr, "Error, ix = %d out of bounds in line %d\n", ix+1, i+1);
	return -1;
      }
      
      if (iy<0 || iy>=*Ny) {
	fprintf (stderr, "Error, iy = %d out of bounds in line %d\n", iy+1, i+1);
	return -1;
      }
      
      /* copy data to result arrays */
      (*albedo)[ix][iy] = value[i][2];
    }

    /* free memory */
    (void) ASCII_free_double (value, rows);
  
    if (!quiet)
      fprintf (stderr, " ... read %d data points from %s\n", 
	       rows-1, filename);

#if HAVE_LIBNETCDF
  }
#endif

  return 0;
}



/*******************************************************/
/* Read data from a 2D umu/phi description file.        */
/*******************************************************/

int read_2D_umu (char *filename, 
		 int *Nx, int *Ny,
		 double *delX, double *delY,
		 double ***umu,
		 double ***phi,
		 int quiet)
{
  int i=0, status=0;
  int ix=0, iy=0;

  double **value=NULL;
  int rows=0, min_columns=0, max_columns=0;

#if HAVE_LIBNETCDF
  size_t start[1] = {0};
 
  size_t tstart[2] = {0,0};
  size_t tcount[2] = {0,0};
 
  size_t n=0;

  int    ncid   =0, idd_nx =0, idd_ny =0;
  int    id_delX=0, id_delY=0, id_type=0;

  float *temp=NULL;
#endif


  /* try to open as netCDF file */
  
#if HAVE_LIBNETCDF
  status = nc_open (filename, NC_NOWRITE, &ncid);
  if (status==NC_NOERR) {

    if (!quiet)
      fprintf (stderr, " ... reading 2D umu data from netCDF file %s\n", filename);

    /* get dimension id for "nx" */
    status = nc_inq_dimid (ncid, "nx", &idd_nx);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading Nx from %s\n", status, filename);
      return status;
    }
    
    /* get dimension length for "nx" */
    status = nc_inq_dimlen (ncid, idd_nx, &n);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading Nx from %s\n", status, filename);
      return status;
    }

    *Nx=n;

    /* get dimension id for "ny" */
    status = nc_inq_dimid (ncid, "ny", &idd_ny);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading Ny from %s\n", status, filename);
      return status;
    }
    
    /* get dimension length for "ny" */
    status = nc_inq_dimlen (ncid, idd_ny, &n);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading Ny from %s\n", status, filename);
      return status;
    }

    *Ny = n;

    /* get variable id for "delX" */
    status = nc_inq_varid (ncid, "dx", &id_delX);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading delX from %s\n", 
	       status, filename);
      return status;
    }

    /* read "delX" */
    status = nc_get_var1_double (ncid, id_delX, start, delX);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading delX from %s\n", 
	       status, filename);
      return status;
    }


    /* get variable id for "delY" */
    status = nc_inq_varid (ncid, "dy", &id_delY);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading dy from %s\n", 
	       status, filename);
      return status;
    }

    /* read "delY" */
    status = nc_get_var1_double (ncid, id_delY, start, delY);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading dy from %s\n", 
	       status, filename);
      return status;
    }


    *delX *= 1000.0;
    *delY *= 1000.0;


    /* allocate memory for 2D umu/phi fields */
    *umu = calloc((size_t) *Nx, sizeof(double *));
    *phi = calloc((size_t) *Nx, sizeof(double *));

    for (ix=0; ix<*Nx; ix++) {
      (*umu)[ix] = calloc((size_t) *Ny, sizeof(double));
      (*phi)[ix] = calloc((size_t) *Ny, sizeof(double));
    }

    /* read data */

    /* get variable id for "umu" */
    status = nc_inq_varid (ncid, "umu", &id_type);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading umu from %s\n", status, filename);
      return status;
    }

    temp = calloc (*Nx, sizeof(float));
    
    /* read type */
    tstart[1] = 0;   /* start with first row          */
    tcount[1] = *Nx; /* read *Nx elements for each iy */

    tcount[0] = 1;  /* read one line at a time */

    for (iy=0; iy<*Ny; iy++) {
      tstart[0] = iy;

      status = nc_get_vara_float (ncid, id_type, tstart, tcount, temp);
      if (status!=NC_NOERR) {
	fprintf (stderr, "Error %d reading umu from %s\n", status, filename);
	return status;
      }

      for (ix=0; ix<*Nx; ix++)
	(*umu)[ix][iy] = temp[ix];
    }

    /* get variable id for "phi" */
    status = nc_inq_varid (ncid, "phi", &id_type);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading phi from %s\n", status, filename);
      return status;
    }

    temp = calloc (*Nx, sizeof(float));
    
    /* read type */
    tstart[1] = 0;   /* start with first row          */
    tcount[1] = *Nx; /* read *Nx elements for each iy */

    tcount[0] = 1;  /* read one line at a time */

    for (iy=0; iy<*Ny; iy++) {
      tstart[0] = iy;

      status = nc_get_vara_float (ncid, id_type, tstart, tcount, temp);
      if (status!=NC_NOERR) {
	fprintf (stderr, "Error %d reading phi from %s\n", status, filename);
	return status;
      }

      for (ix=0; ix<*Nx; ix++)
        /* Switch from DISORT to MYSTIC convention */
	(*phi)[ix][iy] = temp[ix] + 180.0;
    }

    nc_close (ncid);

    free(temp);
    
    if (!quiet)
      fprintf (stderr, " ... read %d data points from %s\n", 
	       (*Nx)*(*Ny), filename);
  }
  else {
    if (!quiet)
      fprintf (stderr, " ... %s not in netCDF format, trying to open as ASCII\n", 
	       filename);
#endif

    /* read file; ASCII_file2double is a waste of memory in this */
    /* case, but it is very convenient because comments and      */
    /* everything are handled correctly.                         */
    
    status = ASCII_file2double (filename,
				&rows, &max_columns, &min_columns, 
				&value);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading file %s\n", status, filename);
      return status;
    }
    
    if (max_columns<4) {
      fprintf (stderr, "Error, found less than four columns in %s\n", 
	       filename);
      return -1;
    }
    
    if (rows<1) {
      if (!quiet)
	fprintf (stderr, "Error, no header in %s\n", filename);
      return -1;
  }
    
    
    /* 1st line, number of cells */
    *Nx = (int) (value[0][0] + 0.5);
    *Ny = (int) (value[0][1] + 0.5);
    
    
    /* 1st line, horizontal distances, convert from km to m */
    *delX = value[0][2] * 1000.0;
    *delY = value[0][3] * 1000.0;
    
    if (!quiet) {
      fprintf (stderr, " ... x-distance: %g\n", *delX);
      fprintf (stderr, " ... y-distance: %g\n", *delY);
    }
    
    /* allocate memory for umu/phi */
    *umu = calloc((size_t) *Nx, sizeof(double *));
    *phi = calloc((size_t) *Nx, sizeof(double *));
    
    for (ix=0; ix<*Nx; ix++) {
      (*umu)[ix] = calloc((size_t) *Ny, sizeof(double));
      (*phi)[ix] = calloc((size_t) *Ny, sizeof(double));
    }
    
    for (i=1; i<rows; i++) {
      ix = (int) (value[i][0] + 0.5) - 1;
      iy = (int) (value[i][1] + 0.5) - 1;
      
      if (ix<0 || ix>=*Nx) {
	fprintf (stderr, "Error, ix = %d out of bounds in line %d\n", ix+1, i+1);
	return -1;
      }
      
      if (iy<0 || iy>=*Ny) {
	fprintf (stderr, "Error, iy = %d out of bounds in line %d\n", iy+1, i+1);
	return -1;
      }
      
      /* copy data to result arrays */
      (*umu)[ix][iy] = value[i][2];
      /* Switch from DISORT to MYSTIC convention */
      (*phi)[ix][iy] = value[i][3] + 180.0; 
    }

    /* free memory */
    (void) ASCII_free_double (value, rows);
  
    if (!quiet)
      fprintf (stderr, " ... read %d data points from %s\n", 
	       rows-1, filename);

#if HAVE_LIBNETCDF
  }
#endif

  return 0;
}


/*******************************************************/
/* Read data from a 2D RPV description file.           */
/*******************************************************/

int read_2D_rpv (char *filename, 
		 char **rpv_labels, int rpv_nlabels,
		 int *Nx, int *Ny,
		 double *delX, double *delY,
		 char ***rpv, int **rpv_count,
		 int *therewereCaMs, int quiet)
{
  int i=0, il=0, status=0;
  int ix=0, iy=0;

  int rows=0, min_columns=0, max_columns=0, max_length=0;

  char ***string=NULL;
  char *dummy=NULL;

#if HAVE_LIBNETCDF
  size_t start[1] = {0};
 
  size_t tstart[2] = {0,0};
  size_t tcount[2] = {0,0};
 
  size_t n=0;

  int    ncid=0; 
  int    idd_nx=0, idd_ny=0;
  int    id_delX=0, id_delY=0, id_type=0;

  unsigned char *temp=NULL;
  char tempstr[4]="";
#endif

  *therewereCaMs=0;

  /* try open as netCDF file */
  
#if HAVE_LIBNETCDF
  status = nc_open (filename, NC_NOWRITE, &ncid);
  if (status==NC_NOERR) {

    if (!quiet)
      fprintf (stderr, " ... reading RPV data from netCDF file %s\n", filename);

    /* get dimension id for "nx" */
    status = nc_inq_dimid (ncid, "nx", &idd_nx);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading Nx from %s\n", status, filename);
      return status;
    }
    
    /* get dimension length for "nx" */
    status = nc_inq_dimlen (ncid, idd_nx, &n);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading Nx from %s\n", status, filename);
      return status;
    }

    *Nx=n;

    /* get dimension id for "ny" */
    status = nc_inq_dimid (ncid, "ny", &idd_ny);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading Ny from %s\n", status, filename);
      return status;
    }
    
    /* get dimension length for "ny" */
    status = nc_inq_dimlen (ncid, idd_ny, &n);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading Ny from %s\n", status, filename);
      return status;
    }

    *Ny=n;

    /* get variable id for "delX" */
    status = nc_inq_varid (ncid, "dx", &id_delX);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading delX from %s\n", 
	       status, filename);
      return status;
    }

    /* read "delX" */
    status = nc_get_var1_double (ncid, id_delX, start, delX);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading delX from %s\n", 
	       status, filename);
      return status;
    }


    /* get variable id for "delY" */
    status = nc_inq_varid (ncid, "dy", &id_delY);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading dy from %s\n", 
	       status, filename);
      return status;
    }

    /* read "delY" */
    status = nc_get_var1_double (ncid, id_delY, start, delY);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading dy from %s\n", 
	       status, filename);
      return status;
    }


    *delX *= 1000.0;
    *delY *= 1000.0;

    /* allocate memory for 2D BRDF field */
    *rpv = calloc((size_t) *Nx, sizeof(char *));
    *rpv_count = calloc((size_t) rpv_nlabels, sizeof(int));

    for (ix=0; ix<*Nx; ix++)
      (*rpv)[ix] = calloc((size_t) *Ny, sizeof(char));

    /* read data */

    /* get variable id for "type" */
    status = nc_inq_varid (ncid, "type", &id_type);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading type from %s\n", status, filename);
      return status;
    }

    temp = calloc (*Nx, sizeof(unsigned char));

    /* read type */
    tstart[1] = 0;   /* start with first x element    */
    tcount[1] = *Nx; /* read *Nx elements for each iy */

    tcount[0] = 1;   /* read one y element at a time  */

    for (iy=0; iy<*Ny; iy++) {
      tstart[0] = iy;

      if (!quiet)
	fprintf (stderr, ".");
      
      status = nc_get_vara_uchar (ncid, id_type, tstart, tcount, temp);
      if (status!=NC_NOERR) {
	fprintf (stderr, "Error %d reading type from %s\n", status, filename);
	return status;
      }

      for (ix=0; ix<*Nx; ix++) {

	sprintf (tempstr, "%d", temp[ix]);
	
	for (il=0; il<rpv_nlabels; il++) {
	  if (!strcmp(tempstr, rpv_labels[il]))
	    break;
	}

	if (il==rpv_nlabels) {
	  fprintf (stderr, "Error, did not find an RPV entry for %s\n", tempstr);
	  return -1;
	}
	
	if (il>UCHAR_MAX) {
	  fprintf (stderr, "Error, RPV index %d larger than %d\n", 
		   il, UCHAR_MAX);
	  return -1;
	}
	  
	(*rpv)[ix][iy] = il; 
	(*rpv_count)[il]++;
      }
    }
    nc_close (ncid);

    free(temp);
    
    if (!quiet)
      fprintf (stderr, "\n");
  }
  else {
    if (!quiet)
      fprintf (stderr, " ... %s not in netCDF format, trying to open as ASCII\n", 
	       filename);
#endif
    
    /* read file; ASCII_file2double is a waste of memory in this */
    /* case, but it is very convenient because comments and      */
    /* everything are handled correctly.                         */
    
    status = ASCII_checkfile (filename, 
			      &rows,
			      &min_columns,
			      &max_columns,
			      &max_length);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    if (max_columns<4) {
      fprintf (stderr, "Error, found less than four columns in %s\n", 
	       filename);
      return -1;
    }
    
    if (rows<1) {
      if (!quiet)
      fprintf (stderr, "Error, no header in %s\n", filename);
      return -1;
    }
    
    
    /* allocate memory */
    status = ASCII_calloc_string (&string,
				  rows,
				  max_columns,
				  max_length);
    
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory\n", status);
      return status;
    }
    
    
    /* read file to string array */
    status = ASCII_readfile (filename, string);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
  
    /* 1st line, number of cells */
    *Nx = strtol(string[0][0], &dummy, 0);
    *Ny = strtol(string[0][1], &dummy, 0);
    
    
    /* 1st line, horizontal distances, convert from km to m */
    *delX = strtod(string[0][2], &dummy) * 1000.0;
    *delY = strtod(string[0][3], &dummy) * 1000.0;
    
    if (!quiet) {
      fprintf (stderr, " ... x-distance: %g\n", *delX);
      fprintf (stderr, " ... y-distance: %g\n", *delY);
    }
    
    /* allocate memory for albedo */
    *rpv = calloc((size_t) *Nx, sizeof(char *));
    *rpv_count = calloc((size_t) rpv_nlabels, sizeof(int));

    for (ix=0; ix<*Nx; ix++)
      (*rpv)[ix] = calloc((size_t) *Ny, sizeof(char));
    
    
    for (i=1; i<rows; i++) {
      ix = strtol(string[i][0], &dummy, 0) - 1;
      iy = strtol(string[i][1], &dummy, 0) - 1;
      
      if (ix<0 || ix>=*Nx) {
	fprintf (stderr, "Error, ix = %d out of bounds in line %d\n", ix+1, i+1);
	return -1;
      }
      
      if (iy<0 || iy>=*Ny) {
	fprintf (stderr, "Error, iy = %d out of bounds in line %d\n", iy+1, i+1);
	return -1;
      }
      
      /* copy data to result arrays */
      for (il=0; il<rpv_nlabels; il++) {
	if (!strcmp(string[i][2], rpv_labels[il]))
	  break;
      }
      
      if (il==rpv_nlabels) {
	fprintf (stderr, "Error, did not find an RPV entry for %s\n", string[i][2]);
	return -1;
      }
      
      if (il>UCHAR_MAX) {
	fprintf (stderr, "Error, RPV index %d larger than %d\n", 
		 il, UCHAR_MAX);
	return -1;
      }
	  
      (*rpv)[ix][iy] = il; 
      (*rpv_count)[il]++;
    }

    /* free memory */
    (void) ASCII_free_string (string, rows, max_columns);

    if (!quiet)
      fprintf (stderr, " ... read %d data points from %s\n", 
	       rows-1, filename);


    if ((*rpv_count)[0]>0)
      *therewereCaMs=1;

#if HAVE_LIBNETCDF
  }    
#endif

  return 0;
}




/*******************************************************/
/* Read data from a 2D Ambrals BRDF description file.  */
/*******************************************************/

int read_2D_ambrals (char *filename, 
		     int *Nx, int *Ny,
		     double *delX, double *delY,
		     ambrals_struct ***ambrals,
		     int *therewereCaMs,
		     int quiet)
{
  int i=0, status=0;
  int ix=0, iy=0;
  int contains_CaM=0, isCaM=0;

  double **value=NULL;
  int rows=0, min_columns=0, max_columns=0;

  *therewereCaMs=0;

  /* read file; ASCII_file2double is a waste of memory in this */
  /* case, but it is very convenient because comments and      */
  /* everything are handled correctly.                         */

  status = ASCII_file2double (filename,
			      &rows, &max_columns, &min_columns, 
			      &value);
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading file %s\n", status, filename);
    return status;
  }

  if (max_columns<5) {
    fprintf (stderr, "Error, found less than five columns in %s\n", 
	     filename);
    return -1;
  }

  /* if more than 5 columns, the 6th column defines whether CaM or not */
  if (max_columns > 5)
    contains_CaM=1;
  
  if (rows<1) {
    if (!quiet)
      fprintf (stderr, "Error, no header in %s\n", filename);
    return -1;
  }


  /* 1st line, number of cells */
  *Nx = (int) (value[0][0] + 0.5);
  *Ny = (int) (value[0][1] + 0.5);


  /* 1st line, horizontal distances, convert from km to m */
  *delX = value[0][2] * 1000.0;
  *delY = value[0][3] * 1000.0;

  if (!quiet) {
    fprintf (stderr, " ... x-distance: %g\n", *delX);
    fprintf (stderr, " ... y-distance: %g\n", *delY);
  }

  /* allocate memory for albedo */
  *ambrals = calloc((size_t) *Nx, sizeof(ambrals_struct *));

  for (ix=0; ix<*Nx; ix++)
    (*ambrals)[ix] = calloc((size_t) *Ny, sizeof(ambrals_struct));


  for (i=1; i<rows; i++) {
    ix = (int) (value[i][0] + 0.5) - 1;
    iy = (int) (value[i][1] + 0.5) - 1;

    if (ix<0 || ix>=*Nx) {
      fprintf (stderr, "Error, ix = %d out of bounds in line %d\n", ix+1, i+1);
      return -1;
    }

    if (iy<0 || iy>=*Ny) {
      fprintf (stderr, "Error, iy = %d out of bounds in line %d\n", iy+1, i+1);
      return -1;
    }

    if (contains_CaM)
      isCaM = (value[i][5] == 1);

    if (isCaM) {
      /* CaM */
      (*ambrals)[ix][iy].iso = 0.0;
      (*ambrals)[ix][iy].vol = 0.0;
      (*ambrals)[ix][iy].geo = 0.0;
      (*ambrals)[ix][iy].isCaM = 1;
      *therewereCaMs=1;
    }
    else {
      /* copy data to result arrays */
      (*ambrals)[ix][iy].iso = value[i][2];
      (*ambrals)[ix][iy].vol = value[i][3];
      (*ambrals)[ix][iy].geo = value[i][4];
      (*ambrals)[ix][iy].isCaM = 0;
    }
  }

  /* free memory */
  (void) ASCII_free_double (value, rows);
  
  if (!quiet)
    fprintf (stderr, " ... read %d data points from %s\n", 
	     rows-1, filename);

  return 0;
}
