"""--------------------------------------------------------------------
 * $Id: atmosphere_options.py 2765 2012-08-17 06:54:27Z svn-kylling $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2012 by Arve Kylling, Bernhard Mayer,
 *                            Claudia Emde, Robert Buras
 *
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------"""


from option_definition import *

class setup_atmosphere_group():

	group_name = 'Atmosphere'

	def __init__(self):
		atmosphere_documentation = get_atmosphere_documentation()

		altitude = option(
			name='altitude',
			group='atmosphere',
			helpstr='Set the bottom level atmosphere provided in atmosphere_file to be at the given altitude.',
			documentation=atmosphere_documentation['altitude'], 
			tokens=[ addToken('Input.alt.altitude',float, 'NOT_DEFINED_FLOAT', valid_range=[-1e6,1e6]),
				 addToken('Input.alt.altitude_dz', float, default=0.0, valid_range=[0,1e6], optional=True ) ],
			settings=addSetting('Input.alt.source','ALT_DIRECT_INPUT', default='ALT_NOT_DEFINED'), 
			parents=['uvspec'], 
			)

		atmosphere_file = mandatory_option(
			name='atmosphere_file',
			group='atmosphere',
			helpstr='Location of the atmosphere data file.', 
			documentation=atmosphere_documentation['atmosphere_file'], 
			tokens= addToken('Input.atmosphere_filename', str, valid_range=['subarctic_winter', 'subarctic_summer', 'midlatitude_summer', 'midlatitude_winter', 'tropics', 'US-standard', file]),  
			parents=['uvspec'], 
			plot = {'plot_type': '2D',
				'optional_args': {'column_names': (
						"altitude",
						"pressure",
						"temperature",
						"air",
						"ozone",
						"oxygen",
						"water vapour",
						"CO2",
						"NO2",)
						  }
				}
			)

		day_of_year = option(
			name='day_of_year',
			group='atmosphere',
			helpstr='Correction fo sun-earth distance',
			documentation=atmosphere_documentation['day_of_year'], 
			tokens=addToken('Input.UTC.tm_yday', int, default='NOT_DEFINED_INTEGER' ),
			parents=['uvspec'], 
			)

		dens_column= not_yet_lex2py_option(
			name='dens_column',
			group='output',
			helpstr='Set column density of specie', 
			documentation=atmosphere_documentation['dens_column'],
			tokens=addToken("", str),
			parents=['uvspec'],
			)
	
		f11_mixing_ratio = option(
			name='f11_mixing_ratio',
			group='atmosphere',
			helpstr='Mixing ratio of F11.', 
			documentation=atmosphere_documentation['f11_mixing_ratio'],
			tokens=addToken('Input.mixing_ratio[MX_F11]', float, default=0.000268, valid_range=[0,1e6]), 
			parents=['uvspec'],
			)

		f12_mixing_ratio = option(
			name='f12_mixing_ratio',
			group='atmosphere', 
			helpstr='Mixing ratio of F12.', 
			documentation=atmosphere_documentation['f12_mixing_ratio'], 
			tokens=addToken('Input.mixing_ratio[MX_F12]', float, default=0.000503, valid_range=[0,1e6]), 
			parents=['uvspec'],
			)

		f22_mixing_ratio = option(
			name='f22_mixing_ratio',
			group='atmosphere', 
			helpstr='Mixing ratio of F22.', 
			documentation=atmosphere_documentation['f22_mixing_ratio'], 
			tokens=addToken('Input.mixing_ratio[MX_F22]', float, default=0.000105, valid_range=[0,1e6]), 
			parents=['uvspec'],
			)

		phi0 = option(
			name='phi0',
			group='atmosphere',
			helpstr='Solar azimuth angle',
			documentation=atmosphere_documentation['phi0'], 
			tokens=addToken('Input.atm.phi0', float, valid_range=[-360.,360.]), 
			parents=['uvspec'],
			)

		pressure = option(
			name='pressure',
			group='atmosphere', 
			helpstr='Surface pressure',
			documentation=atmosphere_documentation['pressure'], 
			tokens=addToken('Input.pressure', float, default='NOT_DEFINED_FLOAT', valid_range=[0,1e6]),
			parents=['uvspec'],
			)

		rayleigh_depol = option(
			name='rayleigh_depol',
			group='atmosphere',
			helpstr='Rayleigh depolarization factor.', 
			documentation=atmosphere_documentation['rayleigh_depol'], 
			tokens=addToken('Input.rayleigh_depol', float, default='NOT_DEFINED_FLOAT'), 
			parents=['uvspec'],
			)

		rh_file = option(
			name='rh_file',
			group='atmosphere',
			helpstr='Location of relative humidity file.', 
			documentation=atmosphere_documentation['rh_file'], 
			tokens=addToken('Input.atm.filename[MOL_H2O]', file),
			settings=addSetting('Input.atm.unit_profile[MOL_H2O]', 'RH', default= 'CM_3'), 
			parents=['uvspec'],
			)

		sza = option(
			name='sza',
			group='atmosphere',
			helpstr='Solar zenith angle',
			documentation=atmosphere_documentation['sza'],
			tokens=addToken('Input.atm.sza', float, default='NOT_DEFINED_FLOAT', valid_range=[0.,180.]), 
			settings=addSetting('Input.atm.sza_source', 'SZA_DIRECT_INPUT', default='SZA_BY_TIME_AND_LOCATION'), 
			parents=['uvspec'],
			non_parents=['sza_file'],
			)

		sza_file = option(
			name='sza_file', 
			group='atmosphere', 
			helpstr='Location of solar zenith angle file.', 
			documentation=atmosphere_documentation['sza_file'], 
			tokens=addToken('Input.filename[FN_SZA]', file),
			settings=addSetting('Input.atm.sza_source', 'SZA_FROM_SZA_FILE', default='SZA_BY_TIME_AND_LOCATION'), 
			parents=['uvspec'],
			non_parents=['sza'],
			)

		ch4_mixing_ratio = option(
			name='ch4_mixing_ratio', 
			group='atmosphere', 
			helpstr='Mixing ratio of CH4 in ppm.', 
			documentation=atmosphere_documentation['ch4_mixing_ratio'], 
			tokens=addToken('Input.mixing_ratio[MX_CH4]', float, default=1.6, valid_range=[0,1e6]), 
			parents=['uvspec'],
			)

		o2_mixing_ratio = option(
			name='o2_mixing_ratio', 
			group='atmosphere',
			helpstr='Mixing ratio of O2 in ppm.',
			documentation=atmosphere_documentation['o2_mixing_ratio'],
			tokens=addToken('Input.mixing_ratio[MX_O2]',float,valid_range=[0,1e6]), 
			parents=['uvspec'],
			)

		n2o_mixing_ratio = option(
			name='n2o_mixing_ratio',
			group='atmosphere', 
			helpstr='Mixing ratio of N2O in ppm.', 
			documentation=atmosphere_documentation['n2o_mixing_ratio'], 
			tokens=addToken('Input.mixing_ratio[MX_N2O]',float,default=0.28,valid_range=[0,1e6]), 
			parents=['uvspec'],
			)

		h2o_mixing_ratio = option(
			name='h2o_mixing_ratio',
			group='atmosphere', 
			helpstr='Mixing ratio of H2O in ppm.', 
			documentation=atmosphere_documentation['h2o_mixing_ratio'], 
			tokens=addToken('Input.mixing_ratio[MX_H2O]',float,valid_range=[0,1e6]), 
			parents=['uvspec'],
			)

		co2_mixing_ratio = option(
			name='co2_mixing_ratio', 
			group='atmosphere', 
			helpstr='Mixing ratio of CO2 in ppm.', #
			documentation=atmosphere_documentation['co2_mixing_ratio'],
			tokens=addToken('Input.mixing_ratio[MX_CO2]',float,valid_range=[0,1e6]), 
			parents=['uvspec'],
			)

		h2o_precip = option(
			name='h2o_precip',
			group='atmosphere', 
			helpstr='Precipitable water in kg / m^2', 
			documentation=atmosphere_documentation['h2o_precip'], 
			tokens=addToken('Input.precip_water',float, default='NOT_DEFINED_FLOAT',valid_range=[0,1e6]), #
			parents=['uvspec'],
			)

		reverse = option(
			name='reverse', 
			group='atmosphere', 
			helpstr='Atmosphere is turned on the head.',
			documentation=atmosphere_documentation['reverse'],
			settings=addSetting('Input.rte.reverse', 1, default=0), 
			parents=['uvspec'],
			)

		zout_interpolate = option(
			name='zout_interpolate', 
			group='atmosphere', #
			helpstr='Interpolate atmospheric profiles.', 
			documentation=atmosphere_documentation['zout_interpolate'], 
			settings=addSetting('Input.atm.zout_interpolate', 'ZOUT_INTERPOLATE', default='NO_ZOUT_INTERPOLATE'),
			parents=['uvspec'],
			)
	

		self.options = [atmosphere_file, sza, sza_file, phi0, 
				dens_column, altitude, pressure, day_of_year,
				ch4_mixing_ratio, co2_mixing_ratio, 
				f11_mixing_ratio, f12_mixing_ratio, f22_mixing_ratio, 
				h2o_mixing_ratio, h2o_precip, n2o_mixing_ratio, o2_mixing_ratio, 
				rayleigh_depol, reverse, rh_file, zout_interpolate ]

	def __iter__(self):
		return iter(self.options)


def get_atmosphere_documentation():
	return {
		'atmosphere_file' : r'''
			Location of the atmospheric data file. 
			\fcode{
			atmosphere_file file
			}
			The file must have at least three columns containing the altitude,
			pressure, and temperature. Missing profiles are filled with 0 (e.g., if you did not specify 
			the ozone profile, there will be no ozone absorption!), with exception of the air density which 
			is calculated from pressure and temperature. Other traces gases may be set by \code{dens_file}.  
			The columns are interpreted as follows:
			\begin{description}
			\item[1] Altitude above sea level in km
			\item[2] Pressure in hPa
			\item[3] Temperature in K
			\item[4] air density in cm$^{-3}$
			\item[5] Ozone density in cm$^{-3}$
			\item[6] Oxygen density in cm$^{-3}$
			\item[7] Water vapour density in cm$^{-3}$
			\item[8] CO2 density in cm$^{-3}$
			\item[9] NO2 density in cm$^{-3}$
			\end{description}
			The atmosphere is specified
			top-down, that is, the top level is the first line in the file, the bottom
			(surface) level the last line. All properties refer to model \emph{level} z,
			not to model \emph{layer}. It is important that the correct units are 
			used, otherwise unpredictable results are guaranteed.
			Comments start with \code{\#}. Empty lines are ignored. Please note that there 
			is some redundancy: For air as an ideal gas the density $\rho$, can be
			calulated from pressure and temperature, $\rho = p / kT$. \code{uvspec} will check 
			if this relation is fulfilled and will stop if it is not. 
			{\sl libRadtran} provides the six standard atmospheres by \citet{Anderson1986}:
			\begin{description}
			\item[afglt]   Tropical            (\code{tropics})                                           
       			\item[afglms]  Midlatitude Summer  (\code{midlatitude_summer})
			\item[afglmw]  Midlatitude Winter  (\code{midlatitude_winter})
			\item[afglss]  Subarctic Summer    (\code{subarctic_summer})
			\item[afglsw]  Subarctic Winter    (\code{subarctic_winter})
			\item[afglus]  U.S. Standard       (\code{US-standard})                   
			\end{description}
			which may be chosen by for example
			\fcode{
			atmosphere_file tropics
			}
			or by specifying the the full file name. These atmosphere files are found in 
			\file{data/atmmod}.
			If no \code{atmosphere_file} is defined, \code{uvspec} will automatically select one. 
			If the information  \code{time}, \code{latitude} and \code{longitude} are provided in 
			the input file \code{uvspec} will choose from the first 5 files, otherwise it takes 
			the U.S. Standard atmosphere.
		''',
		'rh_file' : r'''
			File that defines a profile of relative humidity. 
			\fcode{
			rh_file file
			}
			If specified, the water vapour 
			profile in \code{atmosphere_file} is over-written. If -1 is specified at a level, the value 
			from \code{atmosphere_file} is used. 
		''',
		'altitude' : r'''
			Set the bottom level in the model atmosphere provided in 
			\code{atmosphere_file} to be at the given altitude above sea level (km).
			\fcode{
			altitude 0.73   \# Altitude of IFU, Garmisch-Partenkirchen 
			}
			The profiles of pressure, temperature, molecular absorbers, 
			ice and water clouds are cut at the specified altitude. 
			The aerosol profile is not affected by \code{altitude} but starts
			right from the model surface. This is a convenient way for the user to calculate the 
			radiation at other altitudes than sealevel. Note that \code{altitude} is very different
			from \code{zout} where the radiation is calculated at an altitude of zout
			above the surface. E.g. to calculate the radiation field 1 km above the surface 
			at a location at 0.73 km above sealevel, one would specify '\code{altitude} 0.73' 
			and '\code{zout} 1.0'.
			If an altitude is specified which is below the lowest level in the 
			\code{atmosphere_file}, the atmospheric profiles are extrapolated assuming a constant 
			gradient for temperature and mixing ratios.
			A second optional argument may be given to \code{altitude} as e.g.
			\fcode{
			altitude 0.73 0.5
			}
			
			Here the bottom level will be at 0.73 km and the vertical resolution
			of the model atmosphere will be redistributed to have a spacing
			between levels specified by the second number, here 0.5 km, starting
			however from 0km. (Levels 0.73, 1., 1.5 ... will be added to the
			original atmosphere grid and optical properties are devided into the
			new layers. In order to use interpolated properties use
			\code{zout_interpolate}.  See verbose output for details.)  Be aware
			that specifying a fine vertical spacing will produce many layers thus
			increasing the computing time. Also the radiative transfer equation
			solvers implemented in Fortran 77 might need to have some array sizes
			increased (see \file{src_f/DISORT.MXD}).
			''',
		'day_of_year' : r'''
			Integer, to correct the calculated radiation quantities for the 
			Sun-Earth distance for the specified Julian day (1-365). 
			\fcode{
			   day_of_year value
			}
			If not specified, the Earth-Sun distance is 1 AU (i.e. equinox
			distance), that is, no correction is applied to the extraterrestrial
			irradiance \code{solar_file}. Alternatively \code{time} may be used
			for that purpose.
			''',
		'dens_column' : r'''
			Set the total column of a density profile. The column is integrated between the 
			user-defined \code{altitude} and TOA (top of atmosphere). The syntax is
			\fcode{
			dens_column species column [unit]
			}
			where \code{species} is one of O3, O2, H2O, CO2, NO2, BRO, OCLO, SO2 or HCHO see also \code{dens_file}.
			The second argument is the total column value, and the optional third argument is the unit, 
			in which the column is given. The unit can be DU (Dobson units) or CM\_2 (molecules/cm$^2$).  
			The default units are DU for O3, and CM\_2 for all other gases. It is possible to have
			several \code{dens_column} commands in the input file (maximum one per species). The following sets
			the NO$_2$ total column to 1.2 DU.
			\fcode{
			dens_column NO2 1.2 DU
			}
			''',
		'pressure' : r'''
			The surface pressure (at the user-defined \code{altitude}) in hPa. 
			\fcode{
			pressure value
			}
			The pressure profile as well as air, O2 and CO2 density profiles 
			are scaled accordingly.''',
		'sza' : r'''
			The solar zenith angle (degrees). 
			\fcode{
			sza value
			}
			The default solar zenith angle is 0.''',
		'sza_file' : r'''
			Location of solar zenith angle file for wavelength-dependent solar
			zenith angle. 
			\fcode{
			sza_file file
			}
			This option is useful if you want to simulate an instrument 
			which scans so slowly that the solar zenith angle may change significantly during 
			the wavelength scan. 
			The file must have two or three columns. Column 1 is the wavelength, 
			in nm, and column 2 the corresponding  solar zenith angle. Optionally
			the third column may contain the corresponding solar azimuth angle. The solar 
			azimuth angle is only needed when calculating radiances. The wavelength 
			grid may be freely set. The solar zenith and azimuth angle will be interpolated to 
			the wavelength grid used for the radiation calculation.
			Comments start with \code{\#}. Empty lines are ignored.''',
		'phi0' : r'''
			Azimuth angle of the sun (0 to 360 degrees). 
			\fcode{phi0 value}
			\begin{itemize}
			\item Sun in the South:    0 degrees
			\item Sun in the West:    90 degrees
			\item Sun in the North: 180 degrees
			\item Sun in the East: 270 degrees
			\end{itemize}
			For all one-dimensional solvers the absolute azimuth does not matter, but only the relative azimuth \code{phi}-\code{phi0}.''',
		'f11_mixing_ratio' : r'''
			The mixing ratio of F11 in ppm (default: 0.000268 ppm).
			\fcode{
			f11_mixing_ratio value
			}''',
		'f12_mixing_ratio' : r'''
			The mixing ratio of F12 in ppm (default: 0.000503 ppm).
			\fcode{
			f12_mixing_ratio value
			}''',
		'f22_mixing_ratio' : r'''
			The mixing ratio of F22 in ppm (default: 0.000105 ppm).
			\fcode{
			f22_mixing_ratio value
			}''',
		'rayleigh_depol' : r'''
			Rayleigh depolarization factor.
			\fcode{
			rayleigh_depol value
			}
			The Rayleigh scattering phase function is
			$p(\mu) = a + b  \mu^2$ where $a = 1.5{(1+\texttt{depol})/(2+\texttt{depol})}$ and  
			$b = 1.5{(1-\texttt{depol})/(2+\texttt{depol})}$. By default the depolarization is calculated
			using the expressions from \citet{Bodhaine1999}.''',
		'ch4_mixing_ratio' : r'''
			The mixing ratio of CH4 in ppm (default: 1.6 ppm).
			\fcode{
			 ch4_mixing_ratio value
			}''',		
		'co2_mixing_ratio' : r'''
			The mixing ratio of CO2 in ppm.
			\fcode{
			   co2_mixing_ratio value
			}
			The  profile is scaled so that the mixing 
			ratio at the user-defined \code{altitude} assumes the specified value.''',		
		'h2o_mixing_ratio' : r'''
			The mixing ratio of H2O in ppm. Scale the profile so that the mixing 
			ratio at the user-define \code{altitude} assumes the specified value.
			\fcode{
			h2o_mixing_ratio value
			}''',		
		'n2o_mixing_ratio' : r'''
			The mixing ratio of N2O in ppm (default: 0.28 ppm).
			\fcode{
			n2o_mixing_ratio value
			}''',		
		'o2_mixing_ratio' : r'''
			The mixing ratio of O2 in ppm.
			\fcode{
			o2_mixing_ratio value
			}
			Scale the profile so that the mixing 
			ratio at the user-defined \code{altitude} assumes the specified value.''',
		'h2o_precip' : r'''
			Precipitable water in kg / m2 (which is approximately 1mm). 
			The water vapor profile is scaled accordingly. The precipitable water 
			is integrated from the user-defined \code{altitude} to TOA (top of atmosphere).
			\fcode{
			h2o_precip value
			}''',
		'reverse' : r'''
			Option for the strong and bold. Reverses the atmospheric input to the radiative
			transfer solvers. That is, the atmosphere is turned on the head.
			Yes, that is actually useful for some purposes. If you think
			you need this contact the author. Otherwise, do not use.''',
		'zout_interpolate' : r'''
			The z-grid of optical properties is determined by the \code{atmosphere_file},
			and, if specified, by other profile files like \code{dens_file}, \code{rh_file}, 
			or \code{refractive_index_file}. 
			Additional levels might be introduced by the \code{zout} 
			option and the second argument of the \code{altitude} option. By default 
			(if \code{zout_interpolate} is not specified) levels introduced 
			by the \code{zout} option will not affect the optical property 
			profiles, that is, the optical properties are constant within the layers specified by 
			the \code{atmosphere_file} and profile files.
			If \code{zout_interpolate} is specified, the atmospheric profiles (tracegases, temperature ...) 
			are interpolated to the levels introduced by \code{zout}, and optical 
			properties are determined from the interpolated atmospheric properties.
			If \code{heating_rate}, \code{rte_solver polradtran}, \code{rte_solver rodents},
			or \code{rte_solver twostrebe}
			is specified, \code{zout_interpolate} 
			will also be automatically activated. 
			\code{zout_interpolate} generally causes smoother variation of the optical properties.'''
		}
