"""--------------------------------------------------------------------
 * $Id: molecular_options.py 2803 2012-09-10 20:17:52Z robert.buras $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2012 by Arve Kylling, Bernhard Mayer,
 *                            Claudia Emde, Robert Buras
 *
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------"""


from option_definition import *

class setup_molecular_group():

	group_name = 'Molecular'

	def __init__(self):
		molecular_documentation = get_molecular_documentation()
		o3_crs = option(
			'o3_crs', #name
			'molecular', #group
			'',	#subgroup
			'Ozone cross section',	#helpstr
			molecular_documentation['o3_crs'],	#documentation
			addToken( 'Input.ozone_spectrum_type', str, valid_range=['Bass_and_Paur', 'Molina', 'Daumont', 'Bogumil'] ),	#tokens
			None,	#settings
			['uvspec'],	#parents
			'',	#non_parents
			''	#childs
			)

		rayleigh_crs = option(
			'rayleigh_crs',	#name
			'molecular', #group
			'',	#subgroup
			'Rayleigh cross section.',  #helpstr
			molecular_documentation['rayleigh_crs'],	#documentation
			addToken( 'Input.rayleigh_crs_type', str, valid_range=['Bodhaine', 'Nicolet', 'Penndorf']),	#tokens
			None,	#settings
			['uvspec'],	#parents
			'',	#non_parents
			''	#childs
			)

		absorption = not_yet_lex2py_option(
			'absorption',	#name
			'molecular', #group
			'',	#subgroup
			'Switch off absorption by individual minor trace gases.',  #helpstr
			molecular_documentation['absorption'],	#documentation
			[ addToken('Input.absorption_gas', 'variable_number_of_lines', valid_range=['O4','N2','CO','SO2','NH3','NO','HNO3'] )],	#tokens
			)

		no_absorption = option(
			'no_absorption',	#name
			'molecular', #group
			'',	#subgroup
			'Switch off absorption.',  #helpstr
			molecular_documentation['no_absorption'],	#documentation
			None,	#tokens
			[ addSetting('Input.absorption', 0, default=1),
			  addSetting('Input.molabs', 'MOLABS_NONE', default='MOLABS_CALC') ],	#settings
			['uvspec'],	#parents
			'',	#non_parents
			''	#childs
			)

		no_molecular_absorption = option(
			'no_molecular_absorption',	#name
			'molecular', #group
			'',	#subgroup
			'Switch off molecular absorption.',  #helpstr
			molecular_documentation['no_molecular_absorption'],	#documentation
			None,	#tokens
			addSetting('Input.molabs', 'MOLABS_NONE', default='MOLABS_CALC'),	#settings
			['uvspec'],	#parents
			'',	#non_parents
			''	#childs
			)

		no_rayleigh = option(
			'no_rayleigh',	#name
			'molecular', #group
			'',	#subgroup
			'Switch off Rayleigh scattering.',  #helpstr
			molecular_documentation['no_rayleigh'],	#documentation
			None,	#tokens
			addSetting('Input.rayleigh', 'RAYLEIGH_NONE', default='RAYLEIGH_CALC'),	#settings
			['uvspec'],	#parents
			'',	#non_parents
			''	#childs
			)

		no_scattering = option(
			'no_scattering',	#name
			'molecular', #group
			'',	#subgroup
			'Switch off scattering.',  #helpstr
			molecular_documentation['no_scattering'],	#documentation
			None,	#tokens
			[ addSetting('Input.scattering', 0, default=1),
			  addSetting('Input.rayleigh', 'RAYLEIGH_NONE', default='RAYLEIGH_CALC') ] , #settings
			['uvspec'],	#parents
			'',	#non_parents
			''	#childs
			)

#	crs_file = option(
#		'crs_file',	#name
#		'molecular', #group
#		'',	#subgroup
#		'',  #helpstr
#		molecular_documentation['crs_file'],	#documentation
#		addToken(),	#tokens
#		addSetting(),	#settings
#		'',	#parents
#		'',	#non_parents
#		''	#childs
#	)

		rayleigh_tau_file = option(
			'rayleigh_tau_file',	#name
			'molecular', #group
			'',	#subgroup
			'Location of Rayleigh scattering optical depth file.',  #helpstr
			molecular_documentation['rayleigh_tau_file'],	#documentation
			addToken('Input.filename[FN_RAYLEIGH]', file),	#tokens
			None,	#settings
			['uvspec'],	#parents
			'',	#non_parents
			''	#childs
			)

		molecular_tau_file = option(
			'molecular_tau_file',	#name
			'molecular', #group
			'',	#subgroup
			'Location of molecular absorption optical depth file.',  #helpstr
			molecular_documentation['molecular_tau_file'],	#documentation
			addToken('Input.filename[FN_MOLABS]', file),	#tokens
			None,	#settings
			['uvspec'],	#parents
			'',	#non_parents
			''	#childs
			)

		self.options = [o3_crs, rayleigh_crs, 
			no_absorption, no_molecular_absorption, no_rayleigh, no_scattering, 
			rayleigh_tau_file, molecular_tau_file, absorption]

	def __iter__(self):
		return iter(self.options)


def get_molecular_documentation():
	return {
		'o3_crs'        : r'''
	Choose between the various ozone cross sections by
	\fcode{
	o3_crs type
	}
	where \code{type} is one of
	\begin{description}
	\parameter{Bass_and_Paur} 
	\citet{Bass1985} ozone cross section.
	\parameter{Molina} 
	\citet{Molina1986} ozone cross section.
	\parameter{Daumont} 
	Ozone cross section by \citet{Daumont1992}, \citet{Malicet1995}.
	\parameter{Bogumil} 
	Ozone cross section from \citet{Bogumil2003}.
	\end{description}
	\citet{Molina1986} is default.
	''',
	'rayleigh_crs' : r'''
	Specify the Rayleigh cross section.
	\fcode{
	rayleigh_crs type
	}
	Choose between the following Rayleigh scattering cross sections.
	\begin{description}
	\parameter{Bodhaine} 
	\citet{Bodhaine1999} Rayleigh scattering cross section.
	\parameter{Nicolet} 
	\citet{Nicolet1984} Rayleigh scattering cross section.
	\parameter{Penndorf} 
	\citet{Penndorf1957} Rayleigh scattering cross section.
	\end{description}
	\citet{Bodhaine1999} is default.
	''',
	'absorption' : r'''
	Switch off absorption by individual minor trace gases which are currently only
	included when \code{correlated_k lowtran} is chosen. The syntax is
	\fcode{
	  absorption species on/off
	}
	where species may be one of O4, N2, CO, SO2, NH3, NO, HNO3. By default all
	are switched on. 
	
	This option may also be used to turn on/off absorption by O4 in spectral 
	resolution. It is on by default.
	''',
	'no_absorption' : r'''
	Switch all (molecular, aerosol, cloud, and ice cloud) absorption off. Please 
	note that this option simply sets the absorption optical thickness to 0. 
	If used together with \code{xxx_set_tau} this might be a bit confusing but 
	probably the most logical way. E.g. when using \code{aerosol_default} and 
	\code{aerosol_set_tau 1}, the aerosol optical thickness is set to 1,
	with 0.940539 scattering and 0.059461 absorption. If \code{no_absorption} 
	is added, the absorption optical thickness is set to 0 while the scattering
	optical thickness is preserved at 0.940539 (even though 1 was specified by 
	the user). We find this the most logical solution of the problem because 
	by switching \code{no_absorption} off and on one tests the effect of the 
	absorber in an isolated way, rather than mixing absorption and scattering. 
	The same is true for water and ice clouds. Note, that thermal emission of 
	molecules is also switched off. 
	''',
	'no_molecular_absorption' : r'''
	Switch off molecular absorption.
	''',
	'no_rayleigh' : r'''
	Switch Rayleigh scattering off.
	''',
	'no_scattering' : r'''
	Switch scattering off.
	''',
	'rayleigh_tau_file' : r'''
	Location of Rayleigh scattering optical depth file. 
	\fcode{
	rayleigh_tau_file file
	}
	Usually, the Rayleigh scattering
	cross section is calculated from the air pressure provided in \code{atmosphere_file}
	(scaled with \code{pressure}). Use this parameter only if you really want to specify
	the optical depth directly (e.g. for a model intercomparison). The
	optical thickness profile may be either monochromatic or spectral. 
	The format is exactly the same as for \code{molecular_tau_file}.
	''',
	'molecular_tau_file' : r'''
	Location of molecular absorption optical depth file. 
	\fcode{
	molecular_tau_file file
	}
	Usually, molecular absorption
	is calculated from trace gas concentrations provided in \code{atmosphere_file}
	(scaled with \code{ozone_column}, etc.). Use this option only if you want to specify
	the optical depth directly (e.g. for a model intercomparison) or for a line-by-line
	calculation. If a spectral \code{molecular_tau_file} is specified, the wavelength 
	grid defined there is used as the internal wavelength grid for the radiative transfer
	calculation, if not defined otherwise with \code{transmittance_wl_file}.
	\code{molecular_tau_file} can be either of the following three formats:
	\begin{description}
	\parameter{Monochromatic} 
	Column 1 is the altitude in km 
	%AK: This is probably old stuff prior to redistribute
	%where the altitude grid must be exactly equal 
	%to the altitude grid specified in \code{atmosphere_file}. 
	Column 2 is the absorption optical depth of each layer.
	\parameter{Spectral, ASCII}
	The first line contains the level altitudes in decreasing order; the following lines 
	contain the wavelength [nm] in the first column and then the absorption optical depths
	of each layer. 
	\parameter{Spectral,  netcdf}
	An example is available at the libRadtran homepage, 
	the file \file{UVSPEC.O2A.afglms.cdf} is a line-by-line spectrum of the oxygen A-Band 
	around 760nm, calculated for the mid-latitude summer 
	atmosphere. The advantage of
	netcdf compared to ASCII is that it is much faster to read, and that the file 
	is a self-contained, including data and a description of the variables and arrays.
	It is therefore particularly useful for line-by-line calculations where usually 
	many spectral data points are involved. 
	%netcdf is a common platform independent format; the description, a library to read and 
	%write netcdf including some tools to generate netcdf is available at 
	%http://www.unidata.ucar.edu/packages/netcdf/. A \code{molecular_tau_file} must obey 
	%certain rules; 
	\end{description}
	Comments start with \code{\#}. Empty lines are ignored.
	'''

}
