
"""--------------------------------------------------------------------
 * $Id: option_definition.py 2816 2012-09-11 20:37:58Z svn-jkylling $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2012 by Arve Kylling, Bernhard Mayer,
 *                            Claudia Emde, Robert Buras
 *
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------"""

class option():
	def __init__(self, name=None, group=None, subgroup="",
		     helpstr="", documentation="", tokens=[],
		     settings=[], parents=[], non_parents=[], childs=[],
		     mystic=False, threedmystic=False, plot=None, extra_dependencies=[]):

		assert not name is None
		assert not group is None

		tokens = _checkForDictionary(tokens,'tokens')
		settings = _checkForDictionary(settings,'settings')
		
		if tokens == -1 or settings == -1:
			print '''ERROR: found in definition of %s:\n
			Please check your tokens and settings again''' %(name)

		self.parents = parents
		self.non_parents = non_parents
		self.childs = childs
		self.name = name
		self.help = helpstr
		self.plot = plot
		self.tokens = tokens
		self.threedmystic = threedmystic

		# Please use empty lists
		if self.childs == '': self.childs = []
		if self.non_parents == '': self.non_parents = []

		self.dependencies = list(self.childs + self.non_parents + extra_dependencies)
			
		self.dict = {
			'name'		 : name,		      #as defined in uvspec_lex.l
			'group'          : group,             # e.g. WC
			'subgroup'       : subgroup,          # e.g. microphys_def
			'help'           : helpstr,           # Help string (short), could appear as pop-up in GUI
			'documentation'  : documentation,     # Full documentation 
			'tokens'	 : tokens,	      # Variable in uvspec inp_struct to change
			'settings'       : settings,           # default value for GUI and uvspec_lex.l
			'parents'        : parents,           # (specifies which options must also be defined together with this option)
			# One of the options must also be defined with this options
			'non_parents'    : non_parents,       # specifies which options must not be defined together with this option 
			'childs'    	 : childs,	      # (specifies which options can be defined with this option)
			# Options which will be unlocked when defining this option
			'mystic'	 : mystic,	      # mystic option
			'threedmystic'	 : threedmystic,      # 3D mystic option
			'plot'           : plot,              # Setup plotting for options which should be plotted
			}

	# Pretend to be a dictionary to avoid breaking old code
	def __getitem__(self, *args, **kwargs):
		return self.dict.__getitem__(*args, **kwargs)
	def __setitem__(self, *args, **kwargs):
		return self.dict.__setitem__(*args, **kwargs)
	def __contains__(self, *args, **kwargs):
		return self.dict.__contains__(*args, **kwargs)
	def get(self, *args, **kwargs):
		return self.dict.get(*args, **kwargs)

	def canEnable(self, is_set, get_value):
		"""
		Tells the GUI wether the option should be enabled or disabled.
		Returns True if the option should be enabled and False if it
		should be disabled.

		is_set is a function that returns True if an option is enabled
		and has been edited by the user, else False. It takes one
		argument, the name of an option as a string.

		get_value returns the current value of an option

		This is used to implement the logic in the GUI. If more
		complex logic than the parent, non-parent, children logic is
		needed this function should be overloaded.

		Remember to update the dependency tuple, a tuple of options
		which should be enabled or disabled depending on if this
		option is set.
		"""
		parents = any([is_set(parent) for parent in self.parents]) \
		    or not self.parents
		non_parents = all([not is_set(non_parent) \
					   for non_parent in self.non_parents]) \
					   or not self.non_parents
		return parents and non_parents

	def isMandatory(self, is_set, get_value):
		"""
		Returns True for mandatory options. Similar to canEnable.
		"""
		return False

class SignedFloats(option):
	"""
	Options which can take several floats as argument (e.g. umu, phi, zout)
	Does not currently support settings or tokens, but that should be easy to add.

	I do not know whether the prefix Input.rte is global, it may have to be changed.
	"""
	def writeLex(self):
		return r"""%s{SIGNED_FLOATS}    {\ts = (char *) calloc (strlen(yytext)+1,sizeof(char));\n\tstrcpy (s, yytext);\n\tInput.rte.n%s     = yytext2ntokens(s);\n\tInput.rte.max%s   = Input.rte.n%s;\n\tInput.rte.%s      =\n\t  yytext2floats(yytext, Input.rte.n%s);\n\tfree(s);}\n""" % tuple([self.name]*6)

class mandatory_option(option):
	def isMandatory(self, is_set, get_value):
		return True

class continious_option(option):
	"""
	For options where a change in value can decide if the dependencies
	should be enabled (see ic_properties which triggers ic_habit).
	
	An option with dependencies who uses get_value on this option should be
	instances of this class.
	"""
	continious_update = True

class conditional_enable_option(option):
	def __init__(self, speaker="", enable_values="", **kwargs):
		option.__init__(self, **kwargs)
		self._speaker = speaker
		assert not isinstance(enable_values, basestring), "Missing comma in one-item-tuple?"
		self._enable_values = enable_values
		
	def canEnable(self, is_set, get_value):
		"""
		Remember that self.speaker must be a subclass of
		continious_option.
		"""
		r = option.canEnable(self, is_set, get_value)
		if r and is_set(self._speaker) and  \
			    get_value(self._speaker)[0] in self._enable_values:
			return r
		else:
			return False

class not_yet_lex2py_option(option):
	"""
	Quick fix for options which are currently hard to implement in
	the python-structures.
	"""
	def writeLex(self):
		return ""

class conditional_not_yet_lex2py_option(conditional_enable_option):
	"""
	Quick fix for options which are currently hard to implement in
	the python-structures.
	"""
	def writeLex(self):
		return ""


def addSetting(name, setting, default = None):
	return {
		'name'		: name,		# variable name to set as defined in uvspec_lex.l
		'setting'	: setting, 	# value variable should be set to as in uvspec_lex.l
		'default'	: default 	# default value for uvspec_lex.l
	}

def addToken(name, datatype, default = None, valid_range=None, optional=False):
	return{
		'name'		: name,		# variable name to set as defined in uvspec_lex.l
		'datatype'	: datatype,	# datatype of input token, possible types are float, int, str, file (for GUI)
		'default'	: default,	# default value for GUI AND uvspec_lex.l
		'valid_range'	: valid_range,	# important for GUI and unsigned floats
		'optional'	: optional	# optional argument
	}

def addPlot(plot_type, optional_args={}):
	return {
		'plot_type'     : plot_type,     # One of '2D' or 'map'
		'optional_args' : optional_args # A dictionary of other information required by the specific plot_types. Currently it should be empty for map and contain a list of column_names for 2D.
		}

def _checkForDictionary(argument,tokOrSet):
	
	if tokOrSet == 'settings':
		testDict = addSetting('NAME','SETTING')
	elif tokOrSet == 'tokens':
		testDict = addToken('NAME','TYPE')

	if not argument: return []
	elif type(argument) == list: 
		for arg in argument:
			if type(arg) != dict or sorted(arg.keys()) != sorted(testDict.keys()): return -1
		return argument
				
	elif type(argument) == dict:
		if sorted(argument.keys()) != sorted(testDict.keys()): return -1
		else: return [argument]
	else: return -1

