
"""--------------------------------------------------------------------
 * $Id: writeLex.py 2801 2012-09-08 15:43:02Z bernhard.mayer $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2012 by Arve Kylling, Bernhard Mayer,
 *                            Claudia Emde, Robert Buras
 *
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------"""
from option_definition import SignedFloats

def writeLex():
	import os

	pwd = os.getcwd()
	
	options = loadOptions()					#load python dictionary with input options
	lexParser = loadLexStarter(pwd+'/lex_starter.l')	#load lexStarter.l
	lexParser = insertInputOption(lexParser,options)	#add input options from python dictionary to lex parser
	lexParser = insertDocu(lexParser,options)		#add documentation to python dictionary
	lexParser = insertInitialisations(lexParser,options)	#add default initialisations

	writeUvspecLex(lexParser,pwd+'/uvspec_lex.l')

def loadOptionsHelper():
	"""
	Returns a list of all the groups.
	Used in the GUI.
	"""
	import aerosol_options, ic_options, wc_options, mc_options, atmosphere_options, spectral_options, surface_options, molecular_options, solver_options, output_options

	aer = aerosol_options.setup_aerosol_group()
	atm = atmosphere_options.setup_atmosphere_group()
	mc = mc_options.setup_mc_group()
	spec = spectral_options.setup_spectral_group()
	ic = ic_options.setup_ic_group()
	wc = wc_options.setup_wc_group()
	sur = surface_options.setup_surface_group()
	mol = molecular_options.setup_molecular_group()
	sol = solver_options.setup_solver_group()
	out = output_options.setup_output_group()

	return [atm,spec,aer,wc, ic, mol, sur, sol, mc, out] 

def loadOptions():
	options={}
	
	for group in loadOptionsHelper():
		for option in group:
			options[option.get('name')]=option
	return options	

def loadLexStarter(filename):
	import os

	#read lex file
	pwd = os.getcwd()
	lexstarter = open(filename,'r')
	text = lexstarter.read()
	lexstarter.close()
	return text

def writeUvspecLex(text,filename):
	newlex = open(filename,'w')
	newlex.write(text)
	newlex.close()

def getStr(string):
	if str(string) == 'True': return 'TRUE'
	elif str(string) == 'False': return 'FALSE'
	else: return str(string)
		

def getTypeScan(dtype, valid_range):
	import numpy as np

	if dtype == str or dtype == file:
		return '{SPACE_WORD}'
	elif dtype == int:
		return '{SPACE_INTEGER}'
	elif dtype == long:
		return '{SPACE_FLOAT}'
	elif dtype == float:
		if valid_range and np.sign(valid_range)[0] >= 0:
			return '{SPACE_FLOAT}'
		else:
			return '{SPACE_SIGNED_FLOAT}'

def yytext2type(dtype):
	if dtype == str or dtype == file:
		return 'yytext2string'
	elif dtype == float:
		return 'yytext2float'
	elif dtype == int:
		return 'yytext2int'
	elif dtype == long:
		return '(long int) yytext2double'

def getLexOptions(group):

	inputOption = ''
	
	for name, option in sorted(group.iteritems()):
		# Do something completly different
		if hasattr(option, "writeLex"):
			inputOption += option.writeLex()
		# Do the usual
		else:
			tokens = option.get('tokens')
			nTokens = len(tokens)+1
			optional = ''

			#write input option name
			inputOption += name

			#check for optional arguments and write lex scanner for option
			for tok in tokens:
				inputOption += getTypeScan(tok.get('datatype'),
							   tok.get('valid_range'))
				if tok.get('optional'):
					nTokens = 'ntokens+1'
					optional = '\tntokens = yytext2ntokens(yytext);\n'
					inputOption += '?'
			inputOption += ' {\n'

			#write additional setting
			settings = option.get('settings')
			additionalSetting = ''
			for addSet in settings:
				inputOption += '\t%s = %s;\n' %(addSet.get('name'),
								getStr(addSet.get('setting')))
			inputOption += optional

			#write setting of input arguments
			for itok, tok in enumerate(tokens): 
				if tok.get('optional'):
					inputOption += '\tif ( ntokens == %d )\n\t' %( itok+1 )
				try:
					inputOption += '\t%s = %s(yytext, %d, %s);\n' % ( tok.get('name'),
											  yytext2type(tok.get('datatype')),
											  itok+2, str(nTokens) )
				except Exception, e:
					print 'ERROR found in input option %s :\n\t Datatype %s of variable %s not known!' % (
						option.get('name'), tok.get('datatype'), tok.get('name') )
					print e

			inputOption += '\t}\n'

	return inputOption

def insertInputOption(text,options):
	import re

	#insert input options to lexfile
	groupInput = getLexOptions(options)
	text = re.sub('\n.*PYTHON_OPTIONS.*\n' , '\n%s\n' %(groupInput), text, count=1)
	return text

def insertDocu(text,options):
	import re

	#get new documentation
	docu={}
	for key,val in sorted(options.iteritems()):                  
		docu[key]=val['documentation']
	
	#get whole documentation
	doclist = docu.keys() + re.findall('(?<=option{)\w+',text)
	doclist.sort( key = lambda member: member.lower() )
	
	#insert documentation to lexfile
	for iopt, option in enumerate(reversed(doclist)):
		i=len(doclist)-1-iopt
		if docu.has_key(option) and doclist.count(option) == 1:
			if i>0 and i<len(doclist)-1:
				if text.find('\\ifthreedmystic{\n\\option{%s}' %(doclist[i+1])) > 0: n = text.find('\\ifthreedmystic{\n\\option{%s}' %(doclist[i+1])) 
				elif text.find('\\ifmystic{\n\\option{%s}' %(doclist[i+1])) > 0: n = text.find('\\ifmystic{\n\\option{%s}' %(doclist[i+1])) 
				elif text.find('\\iflidar{\n\\option{%s}' %(doclist[i+1])) > 0: n = text.find('\\iflidar{\n\\option{%s}' %(doclist[i+1]))
				elif text.find('\\undocumented{\n\\option{%s}' %(doclist[i+1])) > 0: n = text.find('\\undocumented{\n\\option{%s}' %(doclist[i+1])) 
				elif text.find('\\option{%s}' %(doclist[i+1])) >0: n = text.find('\\option{%s}' %(doclist[i+1]))
				else: n = text.find('%DOCUMENTATION END')
			else: n = text.find('%DOCUMENTATION END')

			if options[option].get('threedmystic'):
				text = '%s\\ifthreedmystic{\n\\option{%s}%s}\n%s' %(text[:n], option, docu[option], text[n:])
			elif options[option].get('mystic'):
				text = '%s\\ifmystic{\n\\option{%s}%s}\n%s' %(text[:n], option, docu[option], text[n:])
			else:
				text = '%s\\option{%s}%s\n%s' %(text[:n], option, docu[option], text[n:])

	return text

def insertInitialisations(text,options):
	import re

	#get sorted initialisations
	inits = {}
	for val in options.itervalues():
		#get default initialisation of tokens
		for initialize in ['tokens', 'settings']:
			if val.get(initialize):
				for tok in val.get(initialize):
					if tok.get('default') != None: 
						if inits.has_key(tok.get('name')) and inits[tok.get('name')] != tok.get('default'): 

							print 'ERROR: \nvariable %s has different initialisations: %s and %s\n Please check your option definitions again.\n' %(
								tok.get('name'), getStr(inits[tok.get('name')]), getStr(tok.get('default')) )
						else: inits[tok.get('name')] = tok.get('default')
	
	initialisation = ''
	for key in sorted(inits.keys()):
		initialisation += '  %s = %s;\n' %(key,getStr(inits.get(key)) )
	text = re.sub('\n.*PYTHON INITIALISATIONS.*\n' , '\n%s\n' %(initialisation), text, count=1)
	return text

if __name__ == "__main__":
	writeLex()
